package meerkat.bulletinboard;

import com.google.protobuf.Message;
import meerkat.comm.CommunicationException;
import meerkat.crypto.Digest;
import meerkat.crypto.concrete.SHA256Digest;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.rest.Constants;
import meerkat.rest.ProtobufMessageBodyReader;
import meerkat.rest.ProtobufMessageBodyWriter;

import javax.ws.rs.ProcessingException;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.Entity;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.Response;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Callable;

/**
 * Created by Arbel Deutsch Peled on 09-Dec-15.
 *
 * This class implements the actual communication with the Bulletin Board Servers.
 * It is meant to be used in a multi-threaded environment.
 */
//TODO: Maybe make this abstract and inherit from it.
public class BulletinClientWorker implements Callable<BulletinClientJobResult> {

    private final BulletinClientJob job;    // The requested job to be handled

    public BulletinClientWorker(BulletinClientJob job){
        this.job = job;
    }

    // This resource enabled creation of a single Client per thread.
    private static final ThreadLocal<Client> clientLocal =
        new ThreadLocal<Client> () {
            @Override protected Client initialValue() {
                Client client;
                client = ClientBuilder.newClient();
                client.register(ProtobufMessageBodyReader.class);
                client.register(ProtobufMessageBodyWriter.class);

                return client;
            }
        };

    // This resource enables creation of a single Digest per thread.
    private static final ThreadLocal<Digest> digestLocal =
            new ThreadLocal<Digest> () {
                @Override protected Digest initialValue() {
                    Digest digest;
                    digest = new SHA256Digest(); //TODO: Make this generic.

                    return digest;
                }
            };

    /**
     * This method carries out the actual communication with the servers via HTTP Post
     * It accesses the servers according to the job it received and updates said job as it goes
     * The method will only iterate once through the server list, removing servers from the list when they are no longer required
     * In a POST_MESSAGE job: successful post to a server results in removing the server from the list
     * In a GET_REDUNDANCY job: no server is removed from the list and the (absolute) number of servers in which the message was found is returned
     * In a READ_MESSAGES job: successful retrieval from any server terminates the method and returns the received values; The list is not changed
     * @return The original job, modified to fit the current state and the required output (if any) of the operation
     * @throws IllegalArgumentException
     * @throws CommunicationException
     */
    public BulletinClientJobResult call() throws IllegalArgumentException, CommunicationException{

        Client client = clientLocal.get();
        Digest digest = digestLocal.get();

        WebTarget webTarget;
        Response response;

        String requestPath;
        Message msg;

        List<String> serverAddresses = new LinkedList<String>(job.getServerAddresses());

        Message payload = job.getPayload();

        BulletinBoardMessageList msgList;

        int count = 0; // Used to count number of servers which contain the required message in a GET_REDUNDANCY request.

        job.shuffleAddresses(); // This is done to randomize the order of access to servers primarily for READ operations

        // Prepare the request.
        switch(job.getJobType()) {

            case POST_MESSAGE:
                // Make sure the payload is a BulletinBoardMessage
                if (!(payload instanceof BulletinBoardMessage)) {
                    throw new IllegalArgumentException("Cannot post an object that is not an instance of BulletinBoardMessage");
                }

                msg = payload;
                requestPath = Constants.POST_MESSAGE_PATH;
                break;

            case READ_MESSAGES:
                // Make sure the payload is a MessageFilterList
                if (!(payload instanceof MessageFilterList)) {
                    throw new IllegalArgumentException("Read failed: an instance of MessageFilterList is required as payload for a READ_MESSAGES operation");
                }

                msg = payload;
                requestPath = Constants.READ_MESSAGES_PATH;
                break;

            case GET_REDUNDANCY:
                // Make sure the payload is a MessageId
                if (!(payload instanceof MessageID)) {
                    throw new IllegalArgumentException("Cannot search for an object that is not an instance of MessageID");
                }

                requestPath = Constants.READ_MESSAGES_PATH;

                msg = MessageFilterList.newBuilder()
                        .addFilter(MessageFilter.newBuilder()
                                .setType(FilterType.MSG_ID)
                                .setId(((MessageID) payload).getID())
                                .build()
                        ).build();

                break;

            default:
                throw new IllegalArgumentException("Unsupported job type");

        }

        // Iterate through servers

        Iterator<String> addressIterator = serverAddresses.iterator();

        while (addressIterator.hasNext()) {

            // Send request to Server
            String address = addressIterator.next();
            webTarget = client.target(address).path(Constants.BULLETIN_BOARD_SERVER_PATH).path(requestPath);
            response = webTarget.request(Constants.MEDIATYPE_PROTOBUF).post(Entity.entity(msg, Constants.MEDIATYPE_PROTOBUF));

            // Retrieve answer
            switch(job.getJobType()) {

                case POST_MESSAGE:
                    try {

                        response.readEntity(BoolMsg.class); // If a BoolMsg entity is returned: the post was successful
                        addressIterator.remove(); // Post to this server succeeded: remove server from list
                        job.decMinServers();

                    } catch (ProcessingException | IllegalStateException e) {} // Post to this server failed: retry next time
                    finally {
                        response.close();
                    }
                    break;

                case GET_REDUNDANCY:
                    try {
                        msgList = response.readEntity(BulletinBoardMessageList.class); // If a BulletinBoardMessageList is returned: the read was successful

                        if (msgList.getMessageList().size() > 0){ // Message was found in the server.
                            count++;
                        }
                    } catch (ProcessingException | IllegalStateException e) {}  // Read failed: try with next server
                    finally {
                        response.close();
                    }
                    break;

                case READ_MESSAGES:
                    try {
                        msgList = response.readEntity(BulletinBoardMessageList.class); // If a BulletinBoardMessageList is returned: the read was successful
                        return new BulletinClientJobResult(job, msgList); // Return the result
                    } catch (ProcessingException | IllegalStateException e) {} // Read failed: try with next server
                    finally {
                        response.close();
                    }
                    break;

            }

        }

        // Return result (if haven't done so yet)
        switch(job.getJobType()) {

            case POST_MESSAGE:
                // The job now contains the information required to ascertain whether enough server posts have succeeded
                // It will also contain the list of servers in which the post was not successful
                job.updateServerAddresses(serverAddresses);
                return new BulletinClientJobResult(job, null);

            case GET_REDUNDANCY:
                // Return the number of servers in which the message was found
                // The job now contains the list of these servers
                return new BulletinClientJobResult(job, IntMsg.newBuilder().setValue(count).build());

            case READ_MESSAGES:
                // A successful operation would have already returned an output
                // Therefore: no server access was successful
                throw new CommunicationException("Could not access any server");

            default: // This is required for successful compilation
                throw new IllegalArgumentException("Unsupported job type");

        }
    }
}
