package meerkat.bulletinboard.sqlserver;

import com.mysql.jdbc.jdbc2.optional.MysqlDataSource;
import meerkat.bulletinboard.sqlserver.BulletinBoardSQLServer.SQLQueryProvider;
import meerkat.protobuf.BulletinBoardAPI.FilterType;

import javax.sql.DataSource;
import java.util.LinkedList;
import java.util.List;

/**
 * Created by Arbel Deutsch Peled on 09-Dec-15.
 */

public class MySQLQueryProvider implements SQLQueryProvider {

    private String dbAddress;
    private int dbPort;
    private String dbName;
    private String username;
    private String password;

    public MySQLQueryProvider(String dbAddress, int dbPort, String dbName, String username, String password) {
        this.dbAddress = dbAddress;
        this.dbPort = dbPort;
        this.dbName = dbName;
        this.username = username;
        this.password = password;
    }

    @Override
    public String getSQLString(QueryType queryType) throws IllegalArgumentException{

        switch(queryType) {
            case ADD_SIGNATURE:
                return "INSERT IGNORE INTO SignatureTable (EntryNum, SignerId, Signature) VALUES (:EntryNum, :SignerId, :Signature)";
            case CONNECT_TAG:
                return "INSERT IGNORE INTO MsgTagTable (TagId, EntryNum)"
                    + " SELECT TagTable.TagId, :EntryNum AS EntryNum FROM TagTable WHERE Tag = :Tag";
            case FIND_MSG_ID:
                return "SELECT EntryNum From MsgTable WHERE MsgId = :MsgId";
            case GET_MESSAGES:
                return "SELECT MsgTable.EntryNum, MsgTable.Msg FROM MsgTable";
            case GET_SIGNATURES:
                return "SELECT Signature FROM SignatureTable WHERE EntryNum = :EntryNum";
            case INSERT_MSG:
                return "INSERT INTO MsgTable (MsgId, Msg) VALUES(:MsgId, :Msg)";
            case INSERT_NEW_TAG:
                return "INSERT IGNORE INTO TagTable(Tag) VALUES (:Tag)";
           default:
               throw new IllegalArgumentException("Cannot serve a query of type " + queryType);
        }

    }

    @Override
    public String getCondition(FilterType filterType, int serialNum) throws IllegalArgumentException {

        String serialString = Integer.toString(serialNum);

        switch(filterType) {
            case EXACT_ENTRY:
                return "MsgTable.EntryNum = :EntryNum" + serialString;
            case MAX_ENTRY:
                return "MsgTable.EntryNum <= :EntryNum" + serialString;
            case MAX_MESSAGES:
                return "LIMIT :Limit" + serialString;
            case MSG_ID:
                return "MsgTable.MsgId = :MsgId" + serialString;
            case SIGNER_ID:
                return "EXISTS (SELECT 1 FROM SignatureTable"
                        + " WHERE SignatureTable.SignerId = :SignerId" + serialString + " AND SignatureTable.EntryNum = MsgTable.EntryNum)";
            case TAG:
                return "EXISTS (SELECT 1 FROM TagTable"
                        + " INNER JOIN MsgTagTable ON TagTable.TagId = MsgTagTable.TagId"
                        + " WHERE TagTable.Tag = :Tag" + serialString + " AND MsgTagTable.EntryNum = MsgTable.EntryNum)";
            default:
                throw new IllegalArgumentException("Cannot serve a filter of type " + filterType);
        }

    }

    @Override
    public String getConditionParamTypeName(FilterType filterType) throws IllegalArgumentException {

        switch(filterType) {
            case EXACT_ENTRY:   // Go through
            case MAX_ENTRY:     // Go through
            case MAX_MESSAGES:
                return "INT";

            case MSG_ID:        // Go through
            case SIGNER_ID:
                return "TINYBLOB";

            case TAG:
                return "VARCHAR";

            default:
                throw new IllegalArgumentException("Cannot serve a filter of type " + filterType);
        }

    }

    @Override
    public DataSource getDataSource() {
        MysqlDataSource dataSource = new MysqlDataSource();

        dataSource.setServerName(dbAddress);
        dataSource.setPort(dbPort);
        dataSource.setDatabaseName(dbName);
        dataSource.setUser(username);
        dataSource.setPassword(password);

        return dataSource;
    }

    @Override
    public List<String> getSchemaCreationCommands() {
        List<String> list = new LinkedList<String>();

        list.add("CREATE TABLE IF NOT EXISTS MsgTable (EntryNum INT NOT NULL AUTO_INCREMENT PRIMARY KEY, MsgId TINYBLOB, Msg BLOB, UNIQUE(MsgId(50)))");

        list.add("CREATE TABLE IF NOT EXISTS TagTable (TagId INT NOT NULL AUTO_INCREMENT PRIMARY KEY, Tag VARCHAR(50), UNIQUE(Tag))");

        list.add("CREATE TABLE IF NOT EXISTS MsgTagTable (EntryNum INT, TagId INT,"
                + " CONSTRAINT FOREIGN KEY (EntryNum) REFERENCES MsgTable(EntryNum),"
                + " CONSTRAINT FOREIGN KEY (TagId) REFERENCES TagTable(TagId),"
                + " CONSTRAINT UNIQUE (EntryNum, TagID))");

        list.add("CREATE TABLE IF NOT EXISTS SignatureTable (EntryNum INT, SignerId TINYBLOB, Signature TINYBLOB,"
                + " INDEX(SignerId(32)), CONSTRAINT Uni UNIQUE(SignerId(32), EntryNum), CONSTRAINT FOREIGN KEY (EntryNum) REFERENCES MsgTable(EntryNum))");

        return list;
    }

    @Override
    public List<String> getSchemaDeletionCommands() {
        List<String> list = new LinkedList<String>();

        list.add("DROP TABLE IF EXISTS MsgTagTable");
        list.add("DROP TABLE IF EXISTS SignatureTable");
        list.add("DROP TABLE IF EXISTS TagTable");
        list.add("DROP TABLE IF EXISTS MsgTable");

        return list;
    }
}
