package meerkat.voting;

import com.google.protobuf.ByteString;
import meerkat.protobuf.Voting;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.*;
import java.util.concurrent.ArrayBlockingQueue;


public class VotingBoothToyConsoleUI implements VotingBooth.UI, Runnable {
	private BufferedReader m_in;
	private VotingBooth m_vbController;
	private ArrayBlockingQueue<VBMessage> a_queue;
	static private int m_queueSize = 5;
	private Voting.BallotQuestionNew a_questions[];
	private int m_serialNumber;
	private Voting.PlaintextBallot m_plaintextBallot;
	private Voting.EncryptedBallot m_encryptedBallot;
	private Voting.BallotSecrets m_ballotSecrets;
	private final Logger logger;
	private String m_channel;
	private boolean b_isInitialized;
	private boolean b_allowRecastOfAudittedBallot;
	private int m_waitForControllerMillisecTimeout = 100;

	/*
     * constructor for the Console-UI class
     */
	public VotingBoothToyConsoleUI() {
		logger = LoggerFactory.getLogger(VotingBoothToyConsoleUI.class);

		a_queue = new ArrayBlockingQueue<>(m_queueSize, true);
		m_in = new BufferedReader(new InputStreamReader(System.in));

		b_isInitialized = false;
		b_allowRecastOfAudittedBallot = false;
	}

    /*
     * a function for setting behavior of the VotingBoothToy.
     * If it is set to true, then a ballot which was auditted can be printed again for casting.
     * If it is false, then after auditing, the voter has to answer all questions again.
     * Default is false (cannot re-cast an auditted ballot)
     */
	public void setAllowRecastOfAudittedBallot (boolean b) {
		b_allowRecastOfAudittedBallot = b;
	}

	/*
     * initialize the election params
     */
	public void init(Voting.ElectionParams globalParams) {
		m_serialNumber = 0;

		List<Voting.BallotQuestionNew> l_questions =  globalParams.getQuestionList();
		a_questions = new Voting.BallotQuestionNew[l_questions.size()];

		int i = 0;
		for (Voting.BallotQuestionNew q: l_questions) {
			a_questions[i] = q;
			++i;
		}

		b_isInitialized = true;
	}


	public void run() {
		logger.debug("run: initializing the magic!");

        if ((! b_isInitialized) || isHardwareKeyInserted())
        {
			runAdminSetupScenario();
		}

        runVoterScenario();
	}

	/*
     * checks if someone inserted a hardware key.
     * This is used to decided whther to run Voter scenario or Admin-Setup scenario
     */
	private boolean isHardwareKeyInserted() {
		logger.warn("isHardwareKeyInserted: currently always false -> so running Voter scenario");
		return false;
	}

	private void runVoterScenario() {
		while (true) {
			logger.debug("votingBegin: preparing console UI for a new user.");
			// clear history from memory
			eraseEncryption();
			erasePlaintext();
			showWelcomeScreen();
			++m_serialNumber;
			runVotingFlow();
		}
	}

	private void runAdminSetupScenario() {
        if (! isHardwareKeyInserted())
        {
            System.out.println ("Waiting For hardware key to be inserted by admin...");
            showNotInitializedScreen();
        }
        while (! isHardwareKeyInserted())
        {
            try {
                Thread.sleep(100);
            }
            catch (InterruptedException e)
            {
                logger.error ("runAdminSetupScenario: interrupted while waiting for admin hardware key to be inserted");
                throw new RuntimeException("runAdminSetupScenario: interrupted while waiting for admin hardware key to be inserted");
            }
        }
        System.out.println ("runAdminSetupScenario: found hardware key");
		throw new UnsupportedOperationException("runAdminSetupScenario: not yet implemented");
	}

	private String readInputLine() {
		String s;
		try {
			s = this.m_in.readLine();
		} catch (IOException e) {
			logger.error("readInputLine: some error with reading input. " + e);
			return null;
		}

		return s;
	}


	/* (non-Javadoc)
     * @see meerkat.voting.VotingBooth.UI#votingBegin()
     */
	@Override
	public void votingBegin() {
		throw new UnsupportedOperationException("votingBegin: this is actually not a good method for the interface. just run() instead");
	}


	/*
     * This is a welcome screen for the voter
     */
	private void showWelcomeScreen() {
		System.out.println("UI screen: Welcome. Press to start");
		readInputLine();
	}

	/*
     * This is a not-initialized error message screen
     */
	private void showNotInitializedScreen() {
		System.out.println("UI screen: Election params were not yet initialized");
		System.out.println("UI screen: You must reboot the machine with an admin USB stick plugged in");
	}

	/*
     * erase previous encryption from memory
     */
	private void eraseEncryption() {
		//TODO: should we clean memory 'stronger'?
		if (m_encryptedBallot != null) {
			m_encryptedBallot = null;
		}
		if (m_ballotSecrets != null) {
			m_ballotSecrets = null;
		}
	}

	/*
     * erase previous plaintext from memory
     */
	private void erasePlaintext() {
		//TODO: should we clean memory 'stronger'?
		if (m_plaintextBallot != null) {
			m_plaintextBallot = null;
		}
	}

    /*
     * The voter first needs to choose a channel
     */

	private String chooseChannel() {
		System.out.println("UI screen: Choose channel");
		System.out.println("UI screen: currently does nothing. Just type any string you like");
		m_channel = readInputLine();
		return m_channel;
	}

	/*
     * The voting process itself
     */
	private void runVotingFlow() {
		try {
			boolean votingOccured = listQuestionsToUserAndGetAnswers();
			if (votingOccured) {
				printBallotFlow();
			}
		} catch (IOException e) {
			System.err.println("UI debug: IO error");
		}
	}

	private void printBallotFlow () throws IOException{
		logger.warn("UI debug: printBallotFlow currently not multithreaded... will fix it...");

		boolean ended = false;
		while (!ended) {
			sendBallotToControllerForEncryptionAndWaitForResponse ();
			sendBallotToPrinterAndWaitToEnd();
			ended = castOrAudit();
			eraseEncryption();
		}
		erasePlaintext();
	}


	/*
     * This function sends the ballot to the printer
     */
	private void sendBallotToPrinterAndWaitToEnd () {
		logger.warn("UI debug: sendBallotToPrinterAndWaitToEnd: still not multithreaded...");
		System.out.println("UI screen: printing ballot wait for printing to end");
		System.out.println("UI printer: printing ballot");
	}

	/*
     * This function sends the audit to the printer
     */
	private void sendAuditToPrinterAndWaitToEnd () {
		logger.warn("UI debug: sendAuditToPrinterAndWaitToEnd: still not multithreaded...");
		System.out.println("UI printer: printing audit");
	}

	private void showPrintingAuditScreen () {
		System.out.println("UI screen: printing audit wait for printing to end");
	}

	/*
     * checks if the voter wants to cast the ballot or audit
     * returns true if done with this ballot (cast or cancelled)
     * returns false if still needs the current ballot info
     */
	private boolean castOrAudit () throws IOException{
		System.out.println("UI screen: (c)ast or (a)udit?");
		String choice = readInputLine();
        if (null == choice) {
            logger.error("castOrAudit: got null as response");
            throw new IOException();
        }
		if (choice.equals("a") || choice.equals("audit")) {
			logger.debug ("UI debug: voter decided to audit");
			showPrintingAuditScreen();
			sendAuditToPrinterAndWaitToEnd ();

			if (!b_allowRecastOfAudittedBallot) {
				return true;
			}
			System.out.println("Do you still want to keep this vote? Type 'y' for casting/auditing the same vote. Type 'n' tp start from scratch");
			String choice2 = readInputLine();
			if (null == choice2) {
				logger.error("castOrAudit: got null as response");
				throw new IOException ();
			}
			if (choice2.equals("y") || choice.equals("yes")) {
				return false;
			}
			else if (choice2.equals("n") || choice.equals("no")) {
				return true;
			}
			else {
				throw new IOException ();
			}
		}
		else if (choice.equals("c") || choice.equals("cast") || choice.equals("cancel")) {
			System.err.println ("UI debug: voter decided to cast vote.");
			System.out.println("UI screen: wait for printer to finish");
			System.out.println("UI printer: printing empty space till end of page");
			return true;
		}
		else {
			throw new IOException ();
		}
	}

	/*
     * returns true if voting finished successfully
     * false if cancelled in the middle
     */
	private boolean listQuestionsToUserAndGetAnswers() throws IOException {
		Voting.PlaintextBallot.Builder ptbb = Voting.PlaintextBallot.newBuilder();

		ptbb.setSerialNumber(m_serialNumber);

		Voting.BallotAnswer answers[] = new Voting.BallotAnswer[a_questions.length];
		int index = 0;
		while (index < a_questions.length) {
			Voting.BallotQuestionNew q = a_questions[index];
			if (q.getIsMandatory()) {
				throw new UnsupportedOperationException("listQuestionsToUserAndGetAnswers: question " + index + " is marked as mandatory");
			}

			showQuestionOnScreen(index, q);
			System.out.println("UI screen: Enter your answer. You can also type '(b)ack' or '(c)ancel' or '(s)kip");
			String s = readInputLine();
			if (null == s) {
				logger.error("listQuestionsToUserAndGetAnswers: got null as response");
				throw new IOException("");
			}

			if ((s.equals("cancel") || s.equals("c")) || (index == 0 && (s.equals("back") || s.equals("b")))) {
				erasePlaintext();
				return false;
			} else if (s.equals("back") || s.equals("b")) {
				--index;
			} else if (s.equals("skip") || s.equals("s")) {
				answers[index] = translateStringAnswerToProtoBufMessageAnswer("");
				++index;
			} else {
				answers[index] = translateStringAnswerToProtoBufMessageAnswer(s);
				++index;
			}
		}

		ptbb.addAllAnswers(Arrays.asList(answers));
		m_plaintextBallot = ptbb.build();
		return true;
	}


	private Voting.BallotAnswer translateStringAnswerToProtoBufMessageAnswer(String s) {
		Voting.BallotAnswer.Builder bab = Voting.BallotAnswer.newBuilder();
		StringTokenizer st = new StringTokenizer(s);
		while (st.hasMoreTokens()) {
			bab.addAnswer(Integer.parseInt(st.nextToken()));
		}

		return bab.build();
	}

	/*
     * show question on the screen for the voter
     */
	private void showQuestionOnScreen(int i, Voting.BallotQuestionNew q) {

		boolean isText = true;

		if (q.getQuestion().getType() != Voting.UIElementDataType.TEXT
				|| q.getDescription().getType() != Voting.UIElementDataType.TEXT) {
			isText = false;
		}

		for (Voting.UIElement answer : q.getAnswerList()) {
			if (answer.getType() != Voting.UIElementDataType.TEXT) {
				isText = false;
			}
		}

		if (!isText) {
			System.err.println("debug: an element in question " + i + " is not of TEXT type");
			throw new UnsupportedOperationException();
		}


		System.out.println("UI screen: question number " + i);

		System.out.println("Question text: " + bytesToString(q.getQuestion().getData()));
		System.out.println("Description: " + bytesToString(q.getDescription().getData()));
		int answerIndex = 0;
		for (Voting.UIElement answer : q.getAnswerList()) {
			++answerIndex;
			System.out.println("Answer " + answerIndex + ": " + bytesToString(answer.getData()));
		}

	}

	/*
     * Returns the UTF8 decoding of byte-string data
     */
	private static String bytesToString(ByteString data) {
		return data.toStringUtf8();
	}


	/*
     * Registering the controller, so we have an object to send encryption queries to
     */
	public void registerVBController(VotingBooth vb) {
		m_vbController = vb;
	}


	/* (non-Javadoc)
     * @see meerkat.voting.VotingBooth.UI#commitToEncryptedBallot(meerkat.voting.EncryptedBallot)
     */
	@Override
	public void commitToEncryptedBallot(Voting.EncryptedBallot encryptedBallot, Voting.BallotSecrets secrets) {
		try {
			a_queue.put (VBMessage.newControllerResponse(encryptedBallot, secrets));
		}
		catch (InterruptedException e) {
			System.err.println ("Interrupted in VotingBoothToyConsoleUI.commitToEncryptedBallot");
		}
	}


	private void sendBallotToControllerForEncryptionAndWaitForResponse () {

		class TickerTask extends TimerTask {
			private VotingBoothToyConsoleUI m_ui;
			public TickerTask (VotingBoothToyConsoleUI ui) {
				m_ui = ui;
			}

			@Override
			public void run() {
				m_ui.tick();
			}
		}

		eraseEncryption();
		System.out.println("UI screen: Please wait for encryption");
		m_vbController.submitBallot(m_plaintextBallot);
		Timer timer = new Timer();
		timer.scheduleAtFixedRate(new TickerTask(this), new Date(), m_waitForControllerMillisecTimeout);

		while (m_encryptedBallot == null)
		{
			VBMessage msg;
			try {
				msg = a_queue.take();
			}
			catch (InterruptedException e) {
				System.err.println("VotingBoothToyConsoleUI.sendBallotToControllerForEncryptionAndWaitForResponse  interrupted!");
				return;
			}
			if (msg.isEmptyMessage()) {
				System.out.print(".");
			}
			else {
				m_encryptedBallot = msg.getEncryptedBallot();
				if (m_encryptedBallot.getSerialNumber() != m_plaintextBallot.getSerialNumber())
				{
					eraseEncryption();
				}
			}
		}

		logger.debug("sendBallotToControllerForEncryptionAndWaitForResponse: Received EncryptedBallot");
		timer.cancel();
		timer.purge();

	}

}
